from typing import Dict, Optional

def compile(quil: str, quantum_processor_id: str) -> str: ...
def translate(native_quil: str, num_shots: int, quantum_processor_id: str) -> str: ...
def submit(
    program: str, patch_values: Dict[str, float], quantum_processor_id: str
) -> str: ...
def retrieve_results(job_id: str, quantum_processor_id: str) -> str: ...
